import warnings
from typing import Any, Mapping, Union
import sys

if sys.version_info >= (3, 8):
    from typing import Literal
else:
    from typing_extensions import Literal


import numpy as np
import xarray as xr

from . import load, utils


def identity(x):
    return x


def from_interval_no_checks(x):
    return utils.from_interval(x, check=False)


@xr.register_dataset_accessor("emc3")
class EMC3DatasetAccessor:
    """Additional functions for working with EMC3 data."""

    def __init__(self, ds: xr.Dataset):
        self.data = ds
        self.metadata = ds.attrs.get("metadata")  # None if just grid file

    def __str__(self):
        """
        Get a string representation of the EMC3Dataset.

        Accessed by e.g. print(ds.bout).
        """
        text = (
            "<xemc3.EMC3Dataset>\n"
            + "Contains:\n{}\n".format(str(self.data))
            + "Metadata:\n{}\n".format(str(self.metadata))
        )
        return text

    def _get_crop(self, dims, getDA=False, skip=[]):
        for d in dims:
            if d in skip:
                continue
            try:
                if getDA:
                    yield self.data[f"_{d}_dims"]
                else:
                    yield self.data[f"_{d}_dims"].data
            except KeyError:
                try:
                    if getDA:
                        yield self.data[f"{d}_dims"]
                    else:
                        yield self.data[f"{d}_dims"].data
                except KeyError:
                    yield None

    def _get(self, var):
        """Load a single var."""
        transform = identity
        try:
            dims = self.data[var].dims
        except KeyError as e:
            if var.endswith("_corners"):
                var_ = var[: -len("_corners")] + "_bounds"
                try:
                    dims = self.data[var_].dims
                except KeyError:
                    raise e
                var = var_
                transform = from_interval_no_checks
            else:
                raise
        if "plate_ind" in dims:
            crop = list(self._get_crop(dims, skip=["plate_ind"]))
            ret = []
            for i in range(len(self.data["plate_ind"])):
                slcr = tuple(
                    [slice(None) if j is None else slice(None, j[i]) for j in crop]
                )
                data = self.data.isel(plate_ind=i)
                # coords = {
                #     k: xr.DataArray(
                #         coord.data[slcr], dims=coord.dims, attrs=coord.attrs
                #     )
                #     for k, coord in data.coords.items()
                # }
                data = xr.DataArray(
                    data[var].data[slcr],
                    dims=data[var].dims,
                    attrs=data[var].attrs,
                    # coords=coords,
                )
                ret.append(transform(data))
            return ret
        crop = [slice(None, x) for x in self._get_crop(dims)]
        data = self.data[var][tuple(crop)]
        return transform(data)

    def _set(self, var, data):
        """Set a single variable."""
        transform = identity
        if var.endswith("_corners"):
            var = var[: -len("_corners")] + "_bounds"
            transform = utils.to_interval
        # Maybe also do the cropping? See merge code somewhere
        self.data[var] = transform(data)
        return self

    def get(self, *args):
        """
        Get one or more variables of the dataset.

        The shapes are cropped to include only valid data if data for
        the divertor modules is returened.

        The code also transforms to ``*_corner`` cases, if
        ``*_bounds`` is present. The bounds version is an array of
        shape :math:`n\\times...\\times m\\times 2\\times...\\times 2`
        while the corner version is a :math:`n+1\\times...\\times m+1`
        dimensional array.

        Parameters
        ----------
        args : list of str
            The keys of the data to get

        Returns
        -------
        list of xr.DataArray
            The requested arrays from the dataset
        """
        return [self._get(i) for i in args]

    def __getitem__(self, var):
        """
        Get a variable from the dataset.

        See `get` for the transformations performed.

        Parameters
        ----------
        var : str
            The key to load

        Returns
        -------
        xr.DataArray
            The array from the dataset
        """
        return self._get(var)

    def __setitem__(self, var, data):
        """
        Set a variable in the dataset.

        In case a a var ending in ``*_corners`` is set, the data is
        transformed and the corresponding ``*_bounds`` var is set
        instead.

        Parameters
        ----------
        var : str
            The name of the variable
        data : any
            The data to be set. See the xarray documentation for what
            is accepted.
        """
        return self._set(var, data)

    def _get_alt_name(self, var_name, suffix=""):
        var_suffix = ["_bounds", "", "_plus1"]
        var_prefix = ["", "_plate_", "plate_"]
        for extra_suffix in var_suffix:
            for prefix in var_prefix:
                cur = prefix + var_name + extra_suffix + suffix
                if cur in self.data:
                    return cur
        assert False, f"Didn't find variable for {var_name} coordinate!"

    def iter_plates(self, *, symmetry=False, segments=1):
        """
        Iterate over all plates.

        Repeat with stellerator symmetry if symmetry is given. If
        segments is given, repeat for all segments - assuming a 2*pi/n
        rotational transform symmetry.

        Parameters
        ----------
        symmetry: boolean (optional)
            Iterate over mirrored setup
        segments: int (optional)
            Iterate over all segments, assuming a symmetry of
            n=segments.

        Returns
        -------
        iterator of xr.Dataset
            The plates of the divertor plates
        """
        basevar = {k: self._get_alt_name(k) for k in ("phi", "z")}
        var_list = {k: [] for k in basevar.values()}

        if segments != 1 or symmetry:
            for phioffset in np.linspace(0, 2 * np.pi, segments, endpoint=False):
                for sym in [False, True] if symmetry else [False]:
                    for key in var_list:
                        tmp = self.data[key].copy()
                        if sym:
                            tmp *= -1
                        if phioffset != 0 and basevar["phi"] == key:
                            tmp += phioffset
                        var_list[key].append(tmp)
        else:
            for key in var_list:
                var_list[key].append(self.data[key])
        for i in range(len(next(iter(var_list.values())))):
            ds = self.data.copy()
            dims = ds.dims
            crop = self._get_crop(dims, True)
            crop = {
                d: c.data
                for d, c in zip(dims, crop)
                if c is not None and c.dims == ("plate_ind",)
            }
            phid = ds[self._get_alt_name("phi", "_dims")].data
            xd = ds[self._get_alt_name("x", "_dims")].data

            for key in var_list:
                ds[key] = var_list[key][i]
            if len(phid.shape) == 1:
                assert phid.shape == xd.shape
                for j in range(ds.dims["plate_ind"]):
                    yield ds.isel(
                        plate_ind=j, **{k: slice(None, v[j]) for k, v in crop.items()}
                    )
            else:
                for j in range(ds.dims["plate_ind"]):
                    yield ds.isel(plate_ind=j)

    def plot_div(self, index, **kwargs):
        """Plot divertor data."""
        from .plot_3d import divertor

        return divertor(self.data, index, **kwargs)

    def unit(self, var, value=None):
        """Get or set the units of a quantity"""
        if value is not None:
            self.data[var].attrs["units"] = value
        return self.data[var].attrs["units"]

    def to_netcdf(self, fn, complevel=1):
        """
        Write to netcdf file.

        Enables fast compression by default, which helps a lot with
        the sparse data.

        fn: string
            The path of the netcdf file to be written.

        complevel: integer (optional)
            The compression level in the range 1 to 9
        """
        self.data.to_netcdf(
            fn,
            encoding={
                i: {"zlib": True, "complevel": complevel}
                for i in list(self.data) + list(self.data.coords)
            },
        )

    def to_fort(self, keys, fn, kinetic=False):
        """
        Write to text file, using the mappings for plasma or kinetic data.
        """
        if not isinstance(keys, list):
            keys = [keys]
        return load.write_mapped(
            [self.data[k] for k in keys], self.data._plasma_map, fn, kinetic=kinetic
        )

    def from_fort(self, fn, skip_first=0, ignore_broken=False, kinetic=False):
        """
        Read from an EMC3 file, using the mappings for plasma or
        kinetic data.

        Parameters
        ----------
        fn : str
            The name of the file to be read
        skip_first : int (optional)
            For each datablock the first ``skip_first`` lines are
            ignored
        ignore_broken : boolean (optional)
            Incomplete datasets are ignored. This indicates that
            something is wrong, but is the default for EMC3
        kinetic : boolean (optional)
            If kinetic then the data is read using the mapping for
            neutal quantities, otherwise the mapping for plasma
            quantities is assumed

        Returns
        -------
        list of xr.DataArray
            The list of the read quantities
        """
        return load.read_mapped(
            fn, self.data._plasma_map, skip_first, ignore_broken, kinetic
        )

    def plot_rz(self, key, phi, **kwargs):
        """
        Plot a R-z slice in lab coordinates.

        Parameters
        ----------
        key : str or None
            Index of the data to plot. Select None to plot the mesh
            instead.

        phi : float
            Angle at which to plot. As always in radian.

        ax : Axis object (optional)
            Axis object to be used for plotting

        Rmin : float (optional)
            left bound for plot

        Rmax : float (option)
            right bound for plot

        zmin : float (optional)
            lower bound for plot

        zmax : float (optional)
            upper bound for plot

        kwargs: Other arguments to be passed to matplotlib

        Returns
        -------
        matplotlib.collections.QuadMesh
            The return value from matplotlib.pyplot.pcolormesh is returned.
        """
        from . import plot_2d

        return plot_2d.plot_rz(self.data, key, phi, **kwargs)

    def plot_Rz(self, key, phi, **kwargs):
        with warnings.catch_warnings():
            warnings.simplefilter("always", DeprecationWarning)
            warnings.warn(
                "plot_Rz is deprecated. Please switch to plot_rz instead.",
                category=DeprecationWarning,
                stacklevel=2,
            )
        return self.plot_rz(key, phi, **kwargs)

    def plot(self, key, *args, **kw):
        """
        Plot some data.

        In case of data from data from the plates, a 3D plot of the divertor is returned.

        In case of 3D data a volume plot is returend.

        Otherwise xarray is used for plotting. Axes information might be missing.

        See also: plot_rz

        Parameters
        ----------
        key : str
            The key used for plotting

        Returns
        -------
        any
            Return value depends on the specific plotting routine used.
        """
        da = self.data[key]
        if len(da.dims) < 3:
            return da.plot(*args, **kw)
        # For 3D:
        from . import plot_3d

        if "plate_ind" in self.data.dims:
            # assert args == []
            return plot_3d.divertor(self.data, key, *args, **kw)

        init = {}
        for k in "updownsym", "periodicity":
            if k in kw:
                init[k] = kw.pop(k)
        vol = plot_3d.volume(self.data, **init)
        return vol.plot(key, *args, **kw)

    def load(self, path):
        """
        Load EMC3 simulation data from path

        Parameters
        ----------
        path : str
            The directory of the EMC3 simulation data

        Returns
        -------
        xr.Dataset
            The xemc3 dataset with the simulation data
        """
        return load(path)

    def mean_time(self) -> xr.Dataset:
        """
        Average in time.

        Workaround for https://github.com/pydata/xarray/issues/4885
        """
        ds = self.data.copy()
        for k in ds:
            if "time" in ds[k].dims:
                attrs = ds[k].attrs
                ds[k] = ds[k].mean(dim="time")
                ds[k].attrs = attrs
        return ds

    def isel(
        self,
        indexers: Mapping[str, Any] = None,
        drop: bool = False,
        missing_dims: Union[
            Literal["raise"], Literal["warn"], Literal["ignore"]
        ] = "raise",
        **indexers_kwargs: Any,
    ) -> xr.Dataset:
        ds = self.data
        indexers = utils.merge_indexers(indexers, indexers_kwargs)
        mine = {}
        xas = {}
        for k, v in indexers.items():
            if "delta_" + k in ds.dims and k in ds.dims:
                mine[k] = v
            else:
                xas[k] = v
        ds = ds.isel(drop=drop, missing_dims=missing_dims, **xas)
        for k, v in mine.items():
            dk = "delta_" + k
            if v == len(ds[k]):
                ds = ds.isel({k: int(v) - 1, dk: 1})
            elif v == int(v):
                ds = ds.isel({k: int(v), dk: 0})
            else:
                vi = int(v)
                fac = v - vi
                ds_ = (ds.isel({k: vi}) * xr.DataArray([1 - fac, fac], dims=dk)).sum(
                    dim=dk
                )
                for co in ds.coords:
                    if dk in ds.coords[co].dims:
                        ds_[co] = (
                            ds[co].isel({k: vi}) * xr.DataArray([1 - fac, fac], dims=dk)
                        ).sum(dim=dk)
                    else:
                        ds_[co] = ds[co]
                ds = ds_
        return ds

    def sel(
        self,
        indexers: Mapping[str, Any] = None,
        drop: bool = False,
        missing_dims: str = "raise",
        **indexers_kwargs: Any,
    ) -> xr.Dataset:
        ds = self.data
        indexers = utils.merge_indexers(indexers, indexers_kwargs)
        forisel = {}
        for k in indexers.keys():
            val = indexers[k]
            if "delta_" + k in ds.dims and k + "_bounds" in ds:
                assert k in ds.dims
                assert (
                    len(ds[k + "_bounds"].dims) == 2
                ), "Only 1D bounds are currently supported. Maybe try isel."
                dat = ds[k + "_bounds"]
                if dat.dims == (k, "delta_" + k):
                    pass
                elif dat.T.dims == (k, "delta_" + k):
                    dat = dat.T
                else:
                    raise RuntimeError(
                        f"Unexpected dimensions for {k}_bounds - expected {k} and delta_{k} but got {dat.dims}"
                    )
                dat = dat.data
                for i in range(len(dat)):
                    if dat[i, 0] <= val <= dat[i, 1]:
                        break
                else:
                    raise RuntimeError(f"Didn't find {val} in {k}_bounds")
                fac = (val - dat[i, 0]) / (dat[i, 1] - dat[i, 0])
                forisel[k] = i + fac
            else:
                forisel[k] = val
        ds = ds.emc3.isel(forisel)
        assert isinstance(ds, xr.Dataset)
        return ds

    def evaluate_at_xyz(self, x, y, z, *args, **kwargs):
        """
        See evaluate_at_rpz for options. Unlike evaluate_at_rpz the
        coordinates are given here in cartesian coordinates.
        """
        r = np.sqrt(x**2 + y**2)
        phi = np.arctan2(y, x)
        return self.evaluate_at_rpz(r, phi, z, *args, **kwargs)

    def evaluate_at_rpz(
        self,
        r,
        phi,
        z,
        key=None,
        periodicity: int = 5,
        updownsym: bool = True,
        delta_phi: float = None,
        fill_value=None,
        lazy=False,
        progress=False,
    ):
        """
        Evaluate the field key in the dataset at the positions given by
        the array r, phi, z.  If key is None, return the indices to access
        the 3D field and get the appropriate values.

        Parameters
        ----------
        r : array-like
            The (major) radial coordinates to evaluate
        phi : array-like
            The toroidal coordinate
        z : array-like
            The z component
        key : None or str or sequence of str
            If None return the index-coordinates otherwise evaluate the
            specified field in the dataset
        periodicity : int
            The rotational symmetry in toroidal direction
        updownsym : bool
            Whether the data is additionally up-down symmetric with half
            the periodicity.
        delta_phi : None or float
            If not None, delta_phi gives the accuracy of the precision at
            which phi is evaluated. Giving a float enables caching of the
            phi slices, and can speed up computation. Note that it should
            be smaller then the toroidal resolution. For a grid with 1°
            resolution, delta_phi=2 * np.pi / 360 would be the upper
            bound.  None disables caching.
        fill_value : None or any
            If fill_value is None, missing data is initialised as
            np.nan, or as -1 for non-float datatypes. Otherwise
            fill_value is used to set missing data.
        lazy : bool
            Force the loading of the data for key. Defaults to False.
            Can significantly decrease performance, but can decrease
            memory usage.
        progress : bool
            Show the progress of the mapping. Defaults to False.
        """
        from .evaluate_at import _evaluate_get_keys

        at = _evaluate_get_keys(
            self.data, r, phi, z, periodicity, updownsym, delta_phi, progress
        )
        if key is None:
            return at
        if isinstance(key, str):
            keys = [key]
        else:
            keys = key
        ret = xr.Dataset(coords=at.coords)
        fill = at["phi"].data == -1
        filldims = at["phi"].dims
        dofill = np.any(fill)
        for k in keys:
            if not lazy:
                self.data[k].data
            ret[k] = self.data[k].isel(**at)
            if dofill:
                # fillthis = fill if filldims == ret[k].dims else ret[k].data
                assert (
                    ret[k].dims == filldims
                ), f"Got dimensions {ret[k].dims} but expected {filldims} for key {k}"
                if fill_value is None:
                    try:
                        ret[k].data[fill] = np.nan
                    except ValueError:
                        ret[k].data[fill] = -1
                else:
                    ret[k].data[fill] = fill_value
        return ret

    def evaluate_at_diagnostic(self, diag, key=None, num_points=100):
        ret = dict()
        for var in ("los",):
            if var in diag.__dict__:
                for i, dat in enumerate(diag.__dict__[var]):
                    ret[f"{var}_{i}"] = self.evaluate_at_xyz(
                        *dat.get_points(num_points)
                    )
        return ret



import warnings
from typing import Dict

import numpy as np
import xarray as xr
from eudist import PolyMesh  # type: ignore

from . import utils


def _evaluate_get_keys(ds, r, phi, z, periodicity, updownsym, delta_phi, progress):

    with warnings.catch_warnings():
        warnings.filterwarnings("ignore", "invalid value encountered in remainder")
        phi = phi % (np.pi * 2 / periodicity)
    # Get raw data
    dims, shape, coords, (r, phi, z) = get_out_shape(r, phi, z)
    pln = xr.Dataset()
    pln = pln.assign_coords({k: ds[k] for k in ["z_bounds", "R_bounds", "phi_bounds"]})
    lr = len(ds.r)
    lt = len(ds.theta)
    lp = len(ds.phi)
    pln["phi_index"] = "phi", np.arange(lp, dtype=int)
    pln["r_index"] = ("r", "theta"), np.zeros((lr, lt), dtype=int) + np.arange(
        lr, dtype=int
    )[:, None]
    pln["theta_index"] = ("r", "theta"), np.zeros((lr, lt), dtype=int) + np.arange(
        lt, dtype=int
    )
    keys = ["phi_index", "r_index", "theta_index"]

    cache: Dict[int, PolyMesh] = {}
    scache: Dict[int, xr.Dataset] = {}

    n = len(pln.theta)
    outs = [np.empty(shape, dtype=pln[k].data.dtype) for k in keys]
    cid = -1
    assert "delta_phi" in pln.phi_bounds.dims
    assert "phi" in pln.phi_bounds.dims

    rrange = utils.rrange2 if progress else utils.rrange
    for ijk in rrange(shape):
        if not np.isfinite(phi[ijk]):
            for i in range(len(keys)):
                try:
                    outs[i][ijk] = phi[ijk]
                except ValueError:
                    try:
                        outs[i][ijk] = np.nan
                    except ValueError:  # cannot assign nan to integers
                        outs[i][ijk] = -1
            continue

        if updownsym and phi[ijk] > np.pi / periodicity:
            zc = -z[ijk]
            phic = (np.pi * 2 / periodicity) - phi[ijk]
        else:
            zc = z[ijk]
            phic = phi[ijk]

        j = -1
        if delta_phi:
            j = int(round((phic - delta_phi / 2) / delta_phi))

        try:
            mesh = cache[j]
            s = scache[j]
        except KeyError:
            if delta_phi:
                phic = (
                    round((phic - delta_phi / 2) / delta_phi) * delta_phi
                    + delta_phi / 2
                )
                if updownsym and phic > np.pi / periodicity:
                    zc = -zc
                    phic = (np.pi * 2 / periodicity) - phi[ijk]
            s = pln.emc3.sel(phi=phic)
            mesh = PolyMesh(s.emc3["R_corners"].data, s.emc3["z_corners"].data)
            if delta_phi:
                cache[j] = mesh
                scache[j] = s
        cid = mesh.find_cell(np.array([r[ijk], zc]), cid)
        if cid == -1:
            for i in range(len(keys)):
                outs[i][ijk] = -1
        else:
            ij = cid // n, cid % n
            for out, key in zip(outs, keys):
                if len(s[key].dims):
                    out[ijk] = s[key].data[ij]
                else:
                    out[ijk] = s[key].data
    keyout = [k.split("_")[0] for k in keys]

    ret = xr.Dataset(coords=coords)
    for i in range(len(dims)):
        d0 = dims[i]
        cnt = 0
        while dims[i] in keyout:
            dims[i] = f"{d0}_{cnt}"
        if dims[i] != d0 and d0 in ret:
            ret = ret.rename({d0: dims[i]})
    for out, k, ko in zip(outs, keys, keyout):
        ret[k.split("_")[0]] = xr.DataArray(out, dims=dims, attrs=pln[k].attrs)
    return ret


def get_out_shape(*data):
    """
    Convert xarray arrays and plain arrays to same shape and return
    dims, shape and the raw arrays
    """
    if any([isinstance(x, xr.DataArray) for x in data]):
        dims = []
        shape = []
        out = []
        coords = {}
        for d in data:
            if isinstance(d, xr.DataArray):
                for dim in d.dims:
                    if dim in dims:
                        assert len(d[dim]) == shape[dims.index(dim)]
                    else:
                        dims.append(dim)
                        shape.append(len(d[dim]))
                        coords[dim] = d.coords[dim]
            else:
                assert (
                    utils.prod(np.shape(d)) == 1
                ), "Cannot mix `xr.DataArray`s and `np.ndarray`s"
        outzero = xr.DataArray(np.zeros(shape), dims=dims, coords=coords)
        out = [outzero + d for d in data]
        # for o in out:
        #    assert all(o.dims == dims) if len(o.dims) > 1 else o.dims == dims
        out = [d.data for d in out]
        return dims, shape, coords, out
    outzero = np.zeros(1)
    for d in data:
        outzero = outzero * d
    out = [outzero + d for d in data]
    dims = [f"dim_{d}" for d in range(len(outzero.shape))]
    shape = outzero.shape
    return dims, shape, None, out



import itertools
import os
import time
from typing import Any, Callable, Dict, List, Mapping, Optional, Tuple

import numpy as np
import xarray as xr
import warnings

_open_formats: List[Tuple[Callable, str]] = []
try:
    import lzma

    _open_formats += [
        (lzma.open, ".xz"),
        (lzma.open, ".lzma"),
    ]
except ImportError:
    pass
try:
    import gzip

    _open_formats += [(gzip.open, ".gz")]
except ImportError:
    pass


def format_time(secs):
    ret = ""
    out = False
    for name, val in [("m", 60), ("h", 3600), ("d", 60 * 60 * 24)][::-1]:
        if secs >= val:
            out = True
        if out:
            cur = int(secs // val)
            ret += f"{cur:02d}{name}"
            secs -= cur * val
    ret += f"{secs:02.0f}s"
    return ret


class rrange2(object):
    def __init__(self, args, *, update=None):
        if update is not None:
            # This is no forther needed, the update is now based on time.
            warnings.warn(
                "passing the `update` keyword to rrange2 is deprecateed.",
                category=DeprecationWarning,
                stacklevel=2,
            )
        self.args = [int(a) for a in args]
        self.total = 1
        self.t0 = time.time()
        self.tnext = self.t0 + 1
        for a in self.args:
            self.total *= a
        self._has_returned = 0
        return None

    def __iter__(self):
        self.state = [0 for a in self.args]
        if len(self.state) > 0:
            self.state[0] = -1
        self.tstate = -1
        return self

    def __next__(self):
        self.tstate += 1
        self.tnow = time.time()
        if self.tnow >= self.tnext:
            self.tnext = self.tnow + 1
            needed = self.tnow - self.t0
            prog = self.tstate / self.total
            print(
                f"{prog * 100:6.2f} % ... {format_time(needed)}>{format_time(needed / (prog) - needed) if prog else 0}",
                end="\r",
            )
        try:
            self.state[0] += 1
        except IndexError:
            if self.tstate == 0:
                return tuple()
            raise StopIteration
        for i in range(len(self.args)):
            if self.state[i] < self.args[i]:
                return tuple(self.state)
            else:
                self.state[i] = 0
                try:
                    self.state[i + 1] += 1
                except IndexError:
                    print()
                    raise StopIteration


def rrange(args):
    ranges = [range(a) for a in args]
    return itertools.product(*ranges)


def prod(args):
    ret = 1
    for arg in args:
        ret *= arg
    return ret


def to_interval(dims, data=None) -> xr.DataArray:
    """Transforms a N-D i_1+1 x ... x i_N+1 mesh to an
    i_1 x ... x i_N x 2 x ... x 2 mesh of quads
    """
    if isinstance(dims, tuple) and data is None:
        dims, data = dims

    if data is None:
        data = dims
        in_dims = data.dims
        attrs = data.attrs
        data = data.data
    else:
        in_dims = dims
        attrs = {}
    dims = len(data.shape)
    assert dims == len(
        in_dims
    ), f"Data mismatch - {in_dims} as dimensions given, but data is shape {data.shape} (len{len(data.shape)})"
    out_dims = [d for d in in_dims] + ["delta_" + d for d in in_dims]
    ret = np.empty([i - 1 for i in data.shape] + [2] * dims, dtype=data.dtype)
    for ijk in itertools.product(*[range(a - 1) for a in data.shape]):
        tmp = np.array(data[tuple([slice(i, i + 2) for i in ijk])])
        ret[ijk] = tmp
    return xr.DataArray(data=ret, dims=out_dims, attrs=attrs)


def from_interval(data, check=True):
    dims = len(data.shape) // 2
    out_dims = data.dims[:dims]
    attrs = data.attrs
    data = data.data
    ret = np.empty([i + 1 for i in data.shape[:dims]])
    # if dims == 1:
    #     ret[:-1] = data[:, 0]
    #     ret[-1] = data[-1, 1]
    # elif dims == 2:
    #     ret[:-1, :-1] = data[:, :, 0, 0]
    #     ret[-1, :-1] = data[-1, :, 1, 0]
    #     ret[:-1, -1] = data[:, -1, 0, 1]
    #     ret[-1, -1] = data[-1, -1, 1, 1]
    for ijk in itertools.product(*[[0, 1]] * dims):
        first = tuple([[slice(None, -1), -1][i] for i in ijk])
        second = tuple([[slice(None), -1][i] for i in ijk] + [i for i in ijk])
        ret[first] = data[second]
    if check:
        for ijk in itertools.product(*[[0, 1]] * dims):
            first = tuple([[slice(None, -1), slice(1, None)][i] for i in ijk])
            second = tuple([slice(None)] * dims + [i for i in ijk])
            assert (ret[first] == data[second]).all()
    return xr.DataArray(data=ret, dims=out_dims, attrs=attrs)


class timeit:
    def __init__(self, info="%f"):
        self.info = info

    def __enter__(self):
        self.t0 = time.time()

    def __exit__(self, *args):
        print(self.info % (time.time() - self.t0))


T0 = time.time()
last = time.time()


class timeit2:
    def __init__(self, info="%f to %f - %f vs %f"):
        self.info = info

    def __enter__(self):
        self.t0 = time.time()

    def __exit__(self, *args):
        t0 = self.t0
        t1 = time.time()
        global T0, last
        print(self.info % (t0 - T0, t1 - T0, t1 - t0, t0 - last))
        last = t1


def merge_indexers(
    indexers: Optional[Mapping[str, Any]], indexers_kwargs: Dict[str, Any]
) -> Dict[str, Any]:
    if indexers is None:
        return indexers_kwargs.copy()
    for k in indexers.keys():
        if k in indexers_kwargs:
            raise AssertionError(
                f"{k} is provided as kwargs and in indexers - only provide the key once"
            )
    indexers_kwargs = indexers_kwargs.copy()
    indexers_kwargs.update(indexers)
    return indexers_kwargs


def _fft(data):
    from scipy.fftpack import fft  # type: ignore

    fr = fft(data)
    import matplotlib.pyplot as plt  # type: ignore

    plt.plot(np.abs(fr))
    # print(abs(fr.)
    plt.show()
    exit(1)


_open_formats = []
_org_open = open


def open(fn, mode="rt", *args):
    if fn[0] == "~":
        fn = os.environ["HOME"] + fn[1:]
    if mode in ["r", "rb", "rt"]:
        try:
            return _org_open(fn, mode, *args)
        except FileNotFoundError as e:
            for func, ext in _open_formats:
                try:
                    return func(fn + ext, mode, *args)
                except FileNotFoundError:
                    pass
            raise e
    return _org_open(fn, mode, *args)



#!/usr/bin/env python3
import os
import sys
from argparse import ArgumentParser

import xarray as xr

from ..core import dataset
from ..core.load import load_all
from ._common import commonparser, iter_dir


def append_time(read: str, name: str, verbose: bool = False) -> None:
    try:
        os.replace(name + ".nc", name + ".old.nc")
        old = xr.open_dataset(name + ".old.nc")
    except FileNotFoundError:
        old = None
    try:
        if verbose:
            print(f"Loading {read} ...", end="")
            sys.stdout.flush()
        ds = load_all(read)
        if old is not None:
            ds = xr.concat([old, ds], "time", "different")
        if verbose:
            print(" writing ...", end="")
            sys.stdout.flush()
        ds.emc3.to_netcdf(name + ".nc")
    except Exception:
        if old is not None:
            os.replace(name + ".old.nc", name + ".nc")
        raise
    if old is not None:
        os.remove(name + ".old.nc")
    if verbose:
        print(" done")


def parser() -> ArgumentParser:
    return commonparser(
        """
        Load the data from EMC3 simulations and store as netcdf file. The
        data is appended for each simulation to the netcdf file, which
        will be created if it does not yet exists.
        """
    )


def main() -> None:
    args = parser().parse_args()

    for d, n in iter_dir(args):
        append_time(d, n, not args.quiet)


if __name__ == "__main__":
    main()



"""Routines for writing EMC3 files."""

import sys as _sys
import types as _types

# Actual implementation is in load.py
from ..core.load import write_all_fortran as all
from ..core.load import write_locations as coordinates
from ..core.load import write_magnetic_field as magnetic_field
from ..core.load import write_mapped_nice as mapped
from ..core.load import write_plates_mag as plates_mag


class _all(_types.ModuleType):
    def __init__(self):
        super().__init__(__name__)
        self.__dict__.update(_sys.modules[__name__].__dict__)

    def __call__(self, ds, dir):
        return all(ds, dir)


_all.__call__.__doc__ = all.__doc__

_sys.modules[__name__] = _all()



#!/usr/bin/env python3

import os
import sys
from argparse import ArgumentParser

import xarray as xr

import xemc3

long_names = {
    "f_n": "Particle flux",
    "f_E": "Energy flux",
    "avg_n": "Averge density",
    "avg_Te": "Average electron temperature",
    "avg_Ti": "Average ion temperature",
}


def get_name(key):
    try:
        key = long_names[key]
    except KeyError:
        pass
    return key


def get_key(name):
    return {v: k for k, v in long_names.items()}[name]


def plot(cwd, args):
    while cwd[-1] == "/":
        cwd = cwd[:-1]
    if os.path.isdir(cwd):
        cwd += "/"
        plates = xemc3.load.plates(cwd)
    else:
        plates = xr.open_dataset(cwd)
    key = args.key
    if key not in plates:
        key = get_key(key)
    if key == "f_E":
        plates[key].data /= 1e6
        plates[key].attrs["units"] = "M" + plates[key].attrs["units"]

    if args.range == "":
        args.range = ":"
    vmin, vmax = [
        None if s in ["None", ""] else float(s) for s in args.range.split(":")
    ]

    if args.cutoff:
        cutoff = plates["tot_P"].sum() / 100
    else:
        cutoff = None

    segments = 5 if args.plotall else 1

    if args.title:
        title = args.title[0]
    else:
        title = None
    plt1 = plt2 = None
    if args.plotsym and args.plotlower:
        plt1 = cwd + "divertor.png"
        plt2 = cwd + "divertor_zoomed.png"

    plt = plates.emc3.plot_div(
        key,
        symmetry=args.plotsym,
        segments=segments,
        vmax=vmax,
        vmin=vmin,
        power_cutoff=cutoff,
        verbose=not args.quiet,
        only_lower=args.plotlower,
        title=title,
        phi_slices=args.phi_slices,
        path1=plt1,
        path2=plt2,
    )

    if args.gui:
        plt.show()


def parser() -> ArgumentParser:
    parser = ArgumentParser(description="Plot the heatflux on the divertor")
    parser.add_argument(
        "-c",
        "--cutoff",
        action="store_true",
        help="Show only tile with high target power flux",
    )
    parser.add_argument(
        "-s",
        "--plotsym",
        action="store_true",
        help="Plot assuming stellarator symmetry",
    )
    parser.add_argument(
        "-a", "--plotall", action="store_true", help="Plot all 5 segments"
    )
    parser.add_argument(
        "-l", "--plotlower", action="store_true", help="Plot only the lower half"
    )
    parser.add_argument("-q", "--quiet", action="store_true", help="Be less verbose")
    parser.add_argument("-g", "--gui", action="store_true", help="Start the gui")
    parser.add_argument("-r", "--range", default="", help="Set datarange")
    parser.add_argument("-t", "--title", nargs=1, help="Title for plot")
    parser.add_argument(
        "--phi_slices", action="store_true", help="Show where phi in deg is an integer"
    )

    parser.add_argument(
        "-k",
        "--key",
        default="Energy flux",
        type=get_name,
        # choices=[
        #     "Particle flux",
        #     "Energy flux",
        #     "Averge density",
        #     "Average electron temperature",
        #     "Average ion temperature",
        # ],
        help="Data to plot",
    )
    parser.add_argument(
        "-V", "--version", action="version", version=f"xemc3 {xemc3.__version__}"
    )
    parser.add_argument("path", nargs=1, help="Path of data")
    return parser


def main():
    if sys.argv[0] == "mayavi2" or "-x" in sys.argv:
        print("Run with `-g` in python?")
        args = parser().parse_args(sys.argv[3:])
    else:
        args = parser().parse_args()

    for cwd in args.path:
        plot(cwd, args)


if __name__ == "__main__":
    main()



"""Collection of functions which can be used to make a BOUT++ run

    This file is part of boututils.

    boututils is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    boututils is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with boututils.  If not, see <https://www.gnu.org/licenses/>.
"""

import os
import pathlib
import re
import subprocess
from builtins import str
from subprocess import PIPE, STDOUT, Popen, call

if os.name == "nt":
    # Default on Windows
    DEFAULT_MPIRUN = "mpiexec.exe -n"
else:
    DEFAULT_MPIRUN = "mpirun -np"


def getmpirun(default=DEFAULT_MPIRUN):
    """Return environment variable named MPIRUN, if it exists else return
     a default mpirun command

    Parameters
    ----------
    default : str, optional
        An mpirun command to return if ``MPIRUN`` is not set in the environment

    """
    MPIRUN = os.getenv("MPIRUN")

    if MPIRUN is None or MPIRUN == "":
        MPIRUN = default
        print("getmpirun: using the default " + str(default))

    return MPIRUN


def shell(command, pipe=False):
    """Run a shell command

    Parameters
    ----------
    command : list of str
        The command to run, split into (shell) words
    pipe : bool, optional
        Grab the output as text, else just run the command in the
        background

    Returns
    -------
    tuple : (int, str)
        The return code, and either command output if pipe=True else None
    """
    output = None
    status = 0
    if pipe:
        child = Popen(command, stderr=STDOUT, stdout=PIPE, shell=True)
        # This returns a b'string' which is casted to string in
        # python 2. However, as we want to use f.write() in our
        # runtest, we cast this to utf-8 here
        output = child.stdout.read().decode("utf-8", "ignore")
        # Wait for the process to finish. Note that child.wait()
        # would have deadlocked the system as stdout is PIPEd, we
        # therefore use communicate, which in the end also waits for
        # the process to finish
        child.communicate()
        status = child.returncode
    else:
        status = call(command, shell=True)

    return status, output


def determineNumberOfCPUs():
    """Number of virtual or physical CPUs on this system

    i.e. user/real as output by time(1) when called with an optimally
    scaling userspace-only program

    Taken from a post on stackoverflow:
    https://stackoverflow.com/questions/1006289/how-to-find-out-the-number-of-cpus-in-python

    Returns
    -------
    int
        The number of CPUs
    """

    # cpuset
    # cpuset may restrict the number of *available* processors
    try:
        m = re.search(r"(?m)^Cpus_allowed:\s*(.*)$", open("/proc/self/status").read())
        if m:
            res = bin(int(m.group(1).replace(",", ""), 16)).count("1")
            if res > 0:
                return res
    except IOError:
        pass

    # Python 2.6+
    try:
        import multiprocessing

        return multiprocessing.cpu_count()
    except (ImportError, NotImplementedError):
        pass

    # POSIX
    try:
        res = int(os.sysconf("SC_NPROCESSORS_ONLN"))

        if res > 0:
            return res
    except (AttributeError, ValueError):
        pass

    # Windows
    try:
        res = int(os.environ["NUMBER_OF_PROCESSORS"])

        if res > 0:
            return res
    except (KeyError, ValueError):
        pass

    # jython
    try:
        from java.lang import Runtime  # type: ignore

        runtime = Runtime.getRuntime()
        res = runtime.availableProcessors()
        if res > 0:
            return res
    except ImportError:
        pass

    # BSD
    try:
        sysctl = subprocess.Popen(["sysctl", "-n", "hw.ncpu"], stdout=subprocess.PIPE)
        scStdout = sysctl.communicate()[0]
        res = int(scStdout)

        if res > 0:
            return res
    except (OSError, ValueError):
        pass

    # Linux
    try:
        res = open("/proc/cpuinfo").read().count("processor\t:")

        if res > 0:
            return res
    except IOError:
        pass

    # Solaris
    try:
        pseudoDevices = os.listdir("/devices/pseudo/")
        expr = re.compile("^cpuid@[0-9]+$")

        res = 0
        for pd in pseudoDevices:
            if expr.match(pd) is not None:
                res += 1

        if res > 0:
            return res
    except OSError:
        pass

    # Other UNIXes (heuristic)
    try:
        try:
            dmesg = open("/var/run/dmesg.boot").read()
        except IOError:
            dmesgProcess = subprocess.Popen(["dmesg"], stdout=subprocess.PIPE)
            dmesg = dmesgProcess.communicate()[0]

        res = 0
        while "\ncpu" + str(res) + ":" in dmesg:
            res += 1

        if res > 0:
            return res
    except OSError:
        pass

    raise Exception("Can not determine number of CPUs on this system")


def launch(
    command,
    runcmd=None,
    nproc=None,
    mthread=None,
    output=None,
    pipe=False,
    verbose=False,
):
    """Launch parallel MPI jobs

    >>> status = launch(command, nproc, output=None)

    Parameters
    ----------
    command : str
        The command to run
    runcmd : str, optional
        Command for running parallel job; defaults to what getmpirun() returns"
    nproc : int, optional
        Number of processors (default: all available processors)
    mthread : int, optional
        Number of omp threads (default: the value of the
        ``OMP_NUM_THREADS`` environment variable
    output : str, optional
        Name of file to save output to
    pipe : bool, optional
        If True, return the output of the command
    verbose : bool, optional
        Print the full command to be run before running it

    Returns
    -------
    tuple : (int, str)
        The return code, and either command output if pipe=True else None

    """

    if runcmd is None:
        runcmd = getmpirun()

    if nproc is None:
        # Determine number of CPUs on this machine
        nproc = determineNumberOfCPUs()

    cmd = runcmd + " " + str(nproc) + " " + command

    if output is not None:
        cmd = cmd + " > " + output

    if mthread is not None:
        if os.name == "nt":
            # We're on windows, so we have to do it a little different
            cmd = 'cmd /C "set OMP_NUM_THREADS={} && {}"'.format(mthread, cmd)
        else:
            cmd = "OMP_NUM_THREADS={} {}".format(mthread, cmd)

    if verbose:
        print(cmd)

    return shell(cmd, pipe=pipe)


def shell_safe(command, *args, **kwargs):
    """'Safe' version of shell.

    Raises a `RuntimeError` exception if the command is not
    successful

    Parameters
    ----------
    command : str
        The command to run
    *args, **kwargs
        Optional arguments passed to `shell`

    """
    s, out = shell(command, *args, **kwargs)
    if s:
        raise RuntimeError(
            "Run failed with %d.\nCommand was:\n%s\n\n"
            "Output was\n\n%s" % (s, command, out)
        )
    return s, out


def launch_safe(command, *args, **kwargs):
    """'Safe' version of launch.

    Raises an RuntimeError exception if the command is not successful

    Parameters
    ----------
    command : str
        The command to run
    *args, **kwargs
        Optional arguments passed to `shell`

    """
    s, out = launch(command, *args, **kwargs)
    if s:
        raise RuntimeError(
            "Run failed with %d.\nCommand was:\n%s\n\n"
            "Output was\n\n%s" % (s, command, out)
        )
    return s, out


def build_and_log(test):
    """Run make and redirect the output to a log file. Prints input

    On Windows, does nothing because executable should have already
    been built

    """

    if os.name == "nt":
        return

    print("Making {}".format(test))

    if os.path.exists("makefile") or os.path.exists("Makefile"):
        return shell_safe("make > make.log")

    ctest_filename = "CTestTestfile.cmake"
    if not os.path.exists(ctest_filename):
        raise RuntimeError("Could not build: no makefile and no CMake files detected")

    # We're using CMake, but we need to know the target name. If
    # bout_add_integrated_test was used (which it should have been!),
    # then the test name is the same as the target name
    with open(ctest_filename, "r") as f:
        contents = f.read()
    match = re.search("add_test.(.*) ", contents)
    if match is None:
        raise RuntimeError("Using CMake, but could not determine test name")
    test_name = match.group(1).split()[0]

    # Now we need to find the build directory. It'll be the first
    # parent containing CMakeCache.txt
    here = pathlib.Path(".").absolute()
    for parent in here.parents:
        if (parent / "CMakeCache.txt").exists():
            return shell_safe(
                "cmake --build {} --target {} > make.log".format(parent, test_name)
            )

    # We've just looked up the entire directory structure and not
    # found the build directory, this could happen if CMakeCache was
    # deleted, in which case we can't build anyway
    raise RuntimeError("Using CMake, but could not find build directory")
